<?php

declare(strict_types=1);

namespace FiloBlu\Refilo\Helper\Catalog;

use Exception;
use FiloBlu\Refilo\Remote\Entity\Product;
use FiloBlu\Refilo\Remote\Entity\Provider\CategoriesProductProvider;
use FiloBlu\Refilo\Remote\Entity\Provider\PriceProvider;
use FiloBlu\Refilo\Remote\Entity\Provider\StockProvider;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Model\Product\Attribute\Source\Status as AttributeStatus;
use Magento\Catalog\Model\Product\Visibility;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory;
use Magento\CatalogInventory\Helper\Stock;
use Magento\ConfigurableProduct\Model\Product\Type\Configurable;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\DataObject;
use Magento\Framework\EntityManager\MetadataPool;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Store\Model\StoreManagerInterface;

/**
 * Class ProductHelper
 * @package FiloBlu\Refilo\Helper\Catalog
 */
class ProductHelper
{

    /**
     * @var array
     */
    protected $prices = null;

    /**
     * @var array
     */
    protected $stocks = null;

    /**
     * @var array
     */
    protected $categories = null;

    /**
     * @var StockProvider
     */
    private $stockProvider;

    /**
     * @var PriceProvider
     */
    private $priceProvider;

    /**
     * @var StoreManagerInterface
     */
    private $storeManager;
    /**
     * @var CategoriesProductProvider
     */
    private $categoriesProductProvider;
    /**
     * @var CollectionFactory
     */
    private $productCollectionFactory;

    /**
     * @var ResourceConnection
     */
    private $resourceConnection;
    /**
     * @var MetadataPool
     */
    private $metadataPool;
    /**
     * @var Stock
     */
    private $stockHelper;

    /**
     * @var array|null
     */
    private $websiteLookup = [];

    /**
     * ProductHelper constructor.
     * @param StoreManagerInterface $storeManager
     * @param StockProvider $stockProvider
     * @param PriceProvider $priceProvider
     * @param CategoriesProductProvider $categoriesProductProvider
     * @param CollectionFactory $productCollectionFactory
     * @param ResourceConnection $resourceConnection
     * @param MetadataPool $metadataPool
     * @param Stock $stock
     */
    public function __construct(
        StoreManagerInterface $storeManager,
        StockProvider $stockProvider,
        PriceProvider $priceProvider,
        CategoriesProductProvider $categoriesProductProvider,
        CollectionFactory $productCollectionFactory,
        ResourceConnection $resourceConnection,
        MetadataPool $metadataPool,
        Stock $stock
    )
    {
        $this->stockProvider = $stockProvider;
        $this->priceProvider = $priceProvider;
        $this->storeManager = $storeManager;
        $this->categoriesProductProvider = $categoriesProductProvider;
        $this->productCollectionFactory = $productCollectionFactory;
        $this->resourceConnection = $resourceConnection;
        $this->metadataPool = $metadataPool;
        $this->stockHelper = $stock;
    }

    /**
     * @param ProductInterface $product
     * @return mixed
     * @throws NoSuchEntityException
     */
    public function getStockFor(ProductInterface $product)
    {

        $storeId = (int)$product->getStoreId();
        $entityId = $product->getEntityId();

        if ($this->stocks === null || !isset($this->stocks[$storeId])) {
            $store = $this->storeManager->getStore($storeId);
            foreach (iterator_to_array($this->stockProvider->forStore($store)) as $stock) {
                $this->stocks[$storeId][$stock->getId()] = $stock->getData(Product::INVENTORY);
            }
        }

        if (isset($this->stocks[$storeId][$entityId])) {
            return $this->stocks[$storeId][$entityId];
        }

        return [];
    }

    /**
     * @param ProductInterface $product
     * @return mixed
     * @throws NoSuchEntityException
     */
    public function getPricesFor(ProductInterface $product)
    {

        $storeId = (int)$product->getStoreId();
        $entityId = (int)$product->getEntityId();

        if ($this->prices === null || !isset($this->prices[$storeId])) {
            $store = $this->storeManager->getStore($storeId);
            foreach (iterator_to_array($this->priceProvider->forStore($store)) as $price) {
                $this->prices[$storeId][(int)$price->getId()] = $price->getData();
            }

        }


        if (isset($this->prices[$storeId][$entityId])) {
            return $this->prices[$storeId][$entityId];
        }

        return [];
    }

    /**
     * @param ProductInterface $product
     * @return mixed
     * @throws NoSuchEntityException
     */
    public function getFinalPricesFor(ProductInterface $product)
    {

        $storeId = (int)$product->getStoreId();
        $entityId = (int)$product->getEntityId();

        if ($this->prices === null || !isset($this->prices[$storeId])) {
            $store = $this->storeManager->getStore($storeId);
            foreach (iterator_to_array($this->priceProvider->forStore($store)) as $price) {
                foreach ($price->getData() as $key => $value) {
                    if ($key === 'id' || !isset($value['final_price'])) {
                        continue;
                    }
                    $this->prices[$storeId][(int)$price->getId()][$key]['final_price'] = $value['final_price'];
                }
            }

        }

        if (isset($this->prices[$storeId][$entityId])) {
            return $this->prices[$storeId][$entityId];
        }

        return [];
    }

    /**
     * @param ProductInterface $product
     * @return array|mixed
     * @throws NoSuchEntityException
     */
    public function getCategoriesFor(ProductInterface $product)
    {
        $storeId = (int)$product->getStoreId();
        $entityId = (int)$product->getEntityId();

        if ($this->categories === null || !isset($this->categories[$storeId])) {
            $store = $this->storeManager->getStore($storeId);
            foreach (iterator_to_array($this->categoriesProductProvider->forStore($store)) as $category) {
                $this->categories[$storeId][$category->getId()] = $category->getData();
            }
        }

        if (isset($this->categories[$storeId][$entityId])) {
            return $this->categories[$storeId][$entityId];
        }

        return [];
    }

    /**
     * @param \Magento\Catalog\Model\Product $product
     * @param $usedAttributes
     * @return array|DataObject[]
     */
    public function getUsedProducts(\Magento\Catalog\Model\Product $product, $usedAttributes): array
    {
        if ($product->getTypeId() !== Configurable::TYPE_CODE) {
            return [];
        }

        $attributeToSelect = [];
        foreach ($usedAttributes as $attribute) {
            $attributeToSelect[] = $attribute['attribute_code'];
        }

        $childrenCollection = $this->productCollectionFactory->create()
            ->addAttributeToSelect($attributeToSelect)
            ->addStoreFilter($product->getStoreId())
            ->addFieldToFilter('entity_id', [
                'in' => $product->getTypeInstance()->getChildrenIds($product->getId())
            ])
            ->getItems();

        foreach ($childrenCollection as &$child) {
            /** @var \Magento\Catalog\Model\Product $child */
            $this->stockHelper->assignStatusToProduct($child);
        }

        return $childrenCollection;
    }

    /**
     * @param \Magento\Catalog\Model\Product $product
     * @return array
     * @throws Exception
     */
    public function getUsedProductAttributes(\Magento\Catalog\Model\Product $product): array
    {
        $metadata = $this->metadataPool->getMetadata(ProductInterface::class);
        $linkField = $metadata->getLinkField();

        $connection = $this->resourceConnection->getConnection();
        $select = $connection->select()->distinct(
            true
        )->from(
            ['cpsa' => $connection->getTableName('catalog_product_super_attribute')],
            null
        )->joinInner(
            ['cpe' => $connection->getTableName('catalog_product_entity')],
            "cpsa.product_id = cpe.{$linkField}",
            []
        )->joinInner(
            ['ea' => $connection->getTableName('eav_attribute')],
            'cpsa.attribute_id = ea.attribute_id',
            ['ea.attribute_code']
        )->where(
            'cpe.entity_id = :entity_id'
        );

        return $connection->fetchAll($select, ['entity_id' => $product->getEntityId()]) ?? [];
    }

    /**
     * @param $productId
     * @param $websiteId
     * @return bool
     */
    public function productBelongsTo($productId, $websiteId): bool
    {

        $websiteId = (int)$websiteId;

        if (!isset($this->websiteLookup[$websiteId])) {
            $this->websiteLookup[$websiteId] = [];
            $connection = $this->resourceConnection->getConnection();

            $select = $connection->select()
                ->from($connection->getTableName('catalog_product_website'), ['product_id', 'product_id'])
                ->where('website_id = :website_id');

            $this->websiteLookup[$websiteId] = $connection->fetchPairs($select, ['website_id' => $websiteId]);
        }

        return array_key_exists($productId, $this->websiteLookup[$websiteId]);

    }

    /**
     * @param ProductInterface|\Magento\Catalog\Model\Product $magentoProduct
     * @return bool
     */
    public function isProductVisible($magentoProduct): bool
    {
        switch ((int)$magentoProduct->getVisibility()) {
            case Visibility::VISIBILITY_IN_CATALOG:
            case Visibility::VISIBILITY_BOTH:
                return true;
        }

        return false;
    }

    /**
     * @param ProductInterface|\Magento\Catalog\Model\Product $magentoProduct
     * @return bool
     */
    public function isProductVisibleInSearch($magentoProduct): bool
    {
        switch ((int)$magentoProduct->getVisibility()) {
            case Visibility::VISIBILITY_IN_SEARCH:
            case Visibility::VISIBILITY_BOTH:
                return true;
        }

        return false;
    }

    /**
     * @param \Magento\Catalog\Model\Product|ProductInterface $magentoProduct
     * @return string
     */
    public function isEnabled($magentoProduct): string
    {
        return ((int)$magentoProduct->getStatus() === AttributeStatus::STATUS_ENABLED) ? '1' : '0';
    }
}
